/**********************************************************************\
*                                                                      *
*  Title:  Gopt                                                        *
*  Version: 2.0                                                        *
*  Date:  2003-07-14                                                   *
*  Author(s): Tom Vajzovic                                             *
*  Copyright:  Tom Vajzovic                                            *
*  URL:  http://www.ug-home.nottingham.ac.uk/~ppywtcv/gopt/            *
*  Contact:  ppywtcv@nottingham.ac.uk                                  *
*  Description: A GNU compliant command line argument parsing library  *
*                                                                      *
*  You may redistribute and modify this free software under the terms  *
*  of version 2 of the GNU General Public License as published by the  *
*  Free Software Foundation.  http://www.gnu.org/copyleft/gpl.txt      *
*                                                                      *
*  (NOT the Library/Lesser GPL, the full GPL)                          *
*                                                                      *
*  If you redistribute a modified version of this software you must    *
*  do so under a different name.                                       *
*                                                                      *
*  Additionally, the author requests (but does not require) to be      *
*  notified, and that a comment containing the above URL be included.  *
*                                                                      *
\**********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* gopt() return values */
#define GOPT_NOTPRESENT      0
#define GOPT_PRESENT_NOARG   1
#define GOPT_PRESENT_WITHARG 2

/* opt.stat values */
#define GOPT_SHORT_NOARG    0
#define GOPT_SHORT_WITHARG  1
#define GOPT_LONG_NOARG     2
#define GOPT_LONG_WITHARG   3
#define GOPT_NOTHING        4

/*   sorts[] values    */
/* (internal to gopts) */
#define GOPT_COMMAND                  0
#define GOPT_SINGLE_SHORT             1
#define GOPT_SINGLE_SHORT_WITHARG     2
#define GOPT_SINGLE_SHORT_ARGTOFOLLOW 3
#define GOPT_SHORTS_ARG               4
#define GOPT_LONG                     5
#define GOPT_LONG_ARG                 6
#define GOPT_MULTI_SHORT              7
#define GOPT_DOUBLEDASH               8
#define GOPT_ARG                      9

/* a command line option */
typedef struct {

/* long (--option) or short (-o) option   */
/* with or without argument               */
  unsigned char stat;

/* the name of the option:                */
/* pointer to a char for short options    */
/* or a string for long ones              */
  char *name;

/* pointer to the argument of the option  */
/* if it has one (else undefined)         */
  char *val;

} opt;

/* returns a void pointer to an area of memory which must be free()d when you have finished reading options */
/* the void* must be passed as the first argumnent each time you call gopt.                                 */
/* returns NULL and does nothing if out of memory                                                           */
void *gopts(const char* const shortoptswithargs, int* const mainargc, char** const*const mainargv){
  unsigned char expectingoptarg=0,*sorts;
  unsigned int i=1,numopts=0;
  char *eq;
  opt *opts;
#ifdef DEBUG
  fprintf(stderr,"Allocating memory for sorting.\n\n");
#endif
  sorts=(unsigned char*)malloc(*mainargc);
  if(NULL==sorts)
    return NULL;
#ifdef DEBUG
  fprintf(stderr,"The first argument (%s) is the command.\n",(*mainargv)[0]);
#endif
  sorts[0]=GOPT_COMMAND;
  for(;i<(unsigned int)(*mainargc);++i){
#ifdef DEBUG
    fprintf(stderr,"Looking at argument %u (%s).\n",i,(*mainargv)[i]);
#endif
    if((*mainargv)[i][0]=='-'){
#ifdef DEBUG
      fprintf(stderr,"Argument starts with a dash.\n");
#endif
      expectingoptarg=0;
      if((*mainargv)[i][1]=='-'){
#ifdef DEBUG
        fprintf(stderr,"Argument starts with two dashes.\n");
#endif
        if((*mainargv)[i][2]==0){
#ifdef DEBUG
          fprintf(stderr,"Argument is a double dash.\n");
#endif
          sorts[i]=GOPT_DOUBLEDASH;
#ifdef DEBUG
          fprintf(stderr,"All remaining arguments are not options:\n");
#endif
          for(++i;i<(unsigned int)*mainargc;++i){
#ifdef DEBUG
            fprintf(stderr,"Argument %u (%s) is a non-otion argument.\n",i,(*mainargv)[i]);
#endif
            sorts[i]=GOPT_ARG;
          }
          break;
        }else{
#ifdef DEBUG
          fprintf(stderr,"Argument is a long option.\n");
#endif
          if(NULL!=strchr((*mainargv)[i],'=')){
#ifdef DEBUG
            fprintf(stderr,"Argument is a long option with an argument.\n");
#endif
            sorts[i]=GOPT_LONG_ARG;
            ++numopts;
          }else{
#ifdef DEBUG
            fprintf(stderr,"Argument is a long option without an argument.\n");
#endif
            sorts[i]=GOPT_LONG;
            ++numopts;
          }
        }
      }else{
#ifdef DEBUG
        fprintf(stderr,"Argument starts with only one dash.\n");
#endif
        if((*mainargv)[i][1]==0){
#ifdef DEBUG
          fprintf(stderr,"Argument is a single dash non-option argument.\n");
#endif
          sorts[i]=GOPT_ARG;
        }else{
          if((*mainargv)[i][2]==0){
#ifdef DEBUG
            fprintf(stderr,"Argument is a single short option.\n");
#endif
            sorts[i]=GOPT_SINGLE_SHORT;
            ++numopts;
            if(NULL!=strchr(shortoptswithargs,(*mainargv)[i][1])){
#ifdef DEBUG
              fprintf(stderr,"...which may be followed by its argument.\n");
#endif
              expectingoptarg=1;
            }
          }else{
            if(NULL!=strchr(shortoptswithargs,(*mainargv)[i][1])){
#ifdef DEBUG
              fprintf(stderr,"Argument is a single short option, together with its argument.\n");
#endif
              sorts[i]=GOPT_SINGLE_SHORT_WITHARG;
              ++numopts;
            }else{
#ifdef DEBUG
              fprintf(stderr,"Argument is multiple short options.\n");
#endif
              sorts[i]=GOPT_MULTI_SHORT;
              for(eq=(*mainargv)[i]+1;*eq!=0;++eq)
                ++numopts;
            }
          }
        }
      }
    }else{
#ifdef DEBUG
      fprintf(stderr,"Argument doesn't start with a dash.\n");
#endif
      if(expectingoptarg){
#ifdef DEBUG
        fprintf(stderr,"Argument is the argument to the previous option.\n");
#endif
        expectingoptarg=0;
        sorts[i]=GOPT_SHORTS_ARG;
        sorts[i-1]=GOPT_SINGLE_SHORT_ARGTOFOLLOW;
      }else{
#ifdef DEBUG
        fprintf(stderr,"Argument is a non-otion argument.\n");
#endif
        sorts[i]=GOPT_ARG;
      }
    }
  }

#ifdef DEBUG
  fprintf(stderr,"\nTo sumarise I have %d arguments:\n",*mainargc);
  for(i=0;i<*mainargc;++i){
    switch(sorts[i]){
      case GOPT_COMMAND:
        fprintf(stderr,"Argument %u (%s) is the command.\n",i,(*mainargv)[i]);
        break;
      case GOPT_SINGLE_SHORT:
        fprintf(stderr,"Argument %u (%s) is a single short option with no argument.\n",i,(*mainargv)[i]);
        break;
      case GOPT_SINGLE_SHORT_WITHARG:
        fprintf(stderr,"Argument %u (%s) is a single short option with it's argument.\n",i,(*mainargv)[i]);
        break;
      case GOPT_SINGLE_SHORT_ARGTOFOLLOW:
        fprintf(stderr,"Argument %u (%s) is a single short option with it's argument to follow.\n",i,(*mainargv)[i]);
        break;
      case GOPT_SHORTS_ARG:
        fprintf(stderr,"Argument %u (%s) is the argument to the previous option.\n",i,(*mainargv)[i]);
        break;
      case GOPT_LONG:
        fprintf(stderr,"Argument %u (%s) is a long option with no argument.\n",i,(*mainargv)[i]);
        break;
      case GOPT_LONG_ARG:
        fprintf(stderr,"Argument %u (%s) is a long option with an argument.\n",i,(*mainargv)[i]);
        break;
      case GOPT_MULTI_SHORT:
        fprintf(stderr,"Argument %u (%s) is several short options.\n",i,(*mainargv)[i]);
        break;
      case GOPT_DOUBLEDASH:
        fprintf(stderr,"Argument %u (%s) is a double dash.\n",i,(*mainargv)[i]);
        break;
      case GOPT_ARG:
        fprintf(stderr,"Argument %u (%s) is a non-option argument.\n",i,(*mainargv)[i]);
        break;
      default:
        fprintf(stderr,"Something has gone horribly wrong.\n");
    }
  }
  fprintf(stderr,"There are %u options in total.\n\n",numopts);
#endif

  opts=(opt*)malloc((1+numopts)*sizeof(opt));
  if(NULL==opts){
    free(sorts);
    return NULL;
  }

  numopts=0;
  for(i=1;i<(unsigned int)*mainargc;++i){
    switch(sorts[i]){
      case GOPT_SINGLE_SHORT:
#ifdef DEBUG
        fprintf(stderr,"Adding argument %u (%s) as short option %u (%c) with no argument.\n",i,(*mainargv)[i],numopts,*((*mainargv)[i]+1));
#endif
        opts[numopts].stat=GOPT_SHORT_NOARG;
        opts[numopts].name=(*mainargv)[i]+1;
        ++numopts;
        break;
      case GOPT_SINGLE_SHORT_WITHARG:
#ifdef DEBUG
        fprintf(stderr,"Adding argument %u (%s) as short option %u (%c) with argument '%s'.\n",i,(*mainargv)[i],numopts,*((*mainargv)[i]+1),(*mainargv)[i]+2);
#endif
        opts[numopts].stat=GOPT_SHORT_WITHARG;
        opts[numopts].name=(*mainargv)[i]+1;
        opts[numopts].val=(*mainargv)[i]+2;
        ++numopts;
        break;
      case GOPT_SINGLE_SHORT_ARGTOFOLLOW:
#ifdef DEBUG
        fprintf(stderr,"Adding argument %u (%s) as short option %u (%c) with argument '%s'.\n",i,(*mainargv)[i],numopts,*((*mainargv)[i]+1),(*mainargv)[i+1]);
#endif
        opts[numopts].stat=GOPT_SHORT_WITHARG;
        opts[numopts].name=(*mainargv)[i]+1;
        opts[numopts].val=(*mainargv)[i+1];
        ++numopts;
        break;
      case GOPT_LONG:
#ifdef DEBUG
        fprintf(stderr,"Adding argument %u (%s) as long option %u (%s) with no argument.\n",i,(*mainargv)[i],numopts,(*mainargv)[i]+2);
#endif
        opts[numopts].stat=GOPT_LONG_NOARG;
        opts[numopts].name=(*mainargv)[i]+2;
        ++numopts;
        break;
      case GOPT_LONG_ARG:
#ifdef DEBUG
        fprintf(stderr,"Looking for '='.\n");
#endif
        eq=strchr((*mainargv)[i]+2,'=');  /* it must find it, because we already looked for it once */
        *eq=0;
#ifdef DEBUG
        fprintf(stderr,"Adding argument %u (%s) as long option %u (%s) with argument '%s'.\n",i,(*mainargv)[i],numopts,(*mainargv)[i]+2,eq+1);
#endif
        opts[numopts].stat=GOPT_LONG_WITHARG;
        opts[numopts].name=(*mainargv)[i]+2;
        opts[numopts].val=eq+1;
        ++numopts;
        break;
      case GOPT_MULTI_SHORT:
#ifdef DEBUG
        fprintf(stderr,"Adding argument %u (%s) as multiple short options without arguments:\n",i,(*mainargv)[i]);
#endif
        for(eq=(*mainargv)[i]+1;*eq!=0;++eq){
#ifdef DEBUG
          fprintf(stderr,"...option %u is %c.\n",numopts,*eq);
#endif
          opts[numopts].stat=GOPT_SHORT_NOARG;
          opts[numopts].name=eq;
          ++numopts;
        }
    }
  }
#ifdef DEBUG
  fprintf(stderr,"There is no option %u.\n\n",numopts);
#endif
  opts[numopts].stat=GOPT_NOTHING;

  numopts=1;
  for(i=1;i<(unsigned int)*mainargc;++i){
    if(sorts[i]==GOPT_ARG){
#ifdef DEBUG
      fprintf(stderr,"Argument %u (%s) is a non-option argument - moving it to %u in argv.\n",i,(*mainargv)[i],numopts);
#endif
      (*mainargv)[numopts]=(*mainargv)[i];
      ++numopts;
    }
#ifdef DEBUG
    else{
      fprintf(stderr,"Argument %u (%s) is part of an option or a double dash.\nIt will be removed from argv.\n",i,(*mainargv)[i]);
    }
#endif
  }

#ifdef DEBUG
  fprintf(stderr,"\nFreeing sorting memory.\n");
#endif
  free(sorts);

#ifdef DEBUG
  fprintf(stderr,"Adding NULL pointer as argv[%u].\n",numopts);
#endif
  (*mainargv)[numopts]=NULL;

#ifdef DEBUG
  fprintf(stderr,"Corecting argc to %u.\n\n",numopts);
#endif
  *mainargc=numopts;
#ifdef DEBUG
    fprintf(stderr,"gopts() returning.\n");
#endif
  return (void*)opts;
}

/* returns GOPT_NOTPRESENT if --option (string longname) or  -o (char shortname) were not specified at execution   */
/* returns GOPT_PRESENT_NOARG if the option was specified (in either format) without an argument                   */
/* returns GOPT_PRESENT_WITHARG and loads argval with a pointer to the argument if it was present with an argument */
/* use gopt(opts,0,"option",&ptr) if there is no short option or gopt(opts,'o',"",&ptr) for no long.               */
unsigned char gopt(void *vptr2opts, const char shortname, const char *longname, char **argval){
  unsigned int i=0;
  opt *opts;

  if(!(shortname||*longname)){
#ifdef DEBUG
    fprintf(stderr,"This function was called by an idiot.\n");
#endif
    return GOPT_NOTPRESENT;
  }


#ifdef DEBUG
  if(shortname){
    if(*longname)
      fprintf(stderr,"Looking for long option %s or short option %c.\n",longname,shortname);
    else
      fprintf(stderr,"Looking for short option %c.\n",shortname);
  }else
    fprintf(stderr,"Looking for long option %s.\n",longname);
#endif

  opts=(opt*)vptr2opts;
  for(;;++i){
#ifdef DEBUG
    fprintf(stderr,"Testing option %u.\n",i);
#endif
    switch(opts[i].stat){
      case GOPT_SHORT_NOARG:
#ifdef DEBUG
        fprintf(stderr,"...it's a short option with no argument.\n");
#endif
        if(*opts[i].name==shortname){
#ifdef DEBUG
          fprintf(stderr,"Match found, returning with no argument.\n");
#endif
          return GOPT_PRESENT_NOARG;
        }
#ifdef DEBUG
        fprintf(stderr,"...it is %c, no match.\n",*opts[i].name);
#endif
        break;
      case GOPT_SHORT_WITHARG:
#ifdef DEBUG
        fprintf(stderr,"...it's a short option with an argument.\n");
#endif
        if(*opts[i].name==shortname){
#ifdef DEBUG
          fprintf(stderr,"Match found, setting argument pointer and returning.\n");
#endif
          *argval=opts[i].val;
          return GOPT_PRESENT_WITHARG;
        }
#ifdef DEBUG
        fprintf(stderr,"...it is %c, no match.\n",*opts[i].name);
#endif
        break;
      case GOPT_LONG_NOARG:
#ifdef DEBUG
        fprintf(stderr,"...it's a long option with no argument.\n");
#endif
        if(!strcmp(opts[i].name,longname)){
#ifdef DEBUG
          fprintf(stderr,"Match found, returning with no argument.\n");
#endif
          return GOPT_PRESENT_NOARG;
        }
#ifdef DEBUG
        fprintf(stderr,"...it is %s, no match.\n",opts[i].name);
#endif
        break;
      case GOPT_LONG_WITHARG:
#ifdef DEBUG
        fprintf(stderr,"...it's a long option with an argument.\n");
#endif
        if(!strcmp(opts[i].name,longname)){
#ifdef DEBUG
          fprintf(stderr,"Match found, setting argument pointer and returning.\n");
#endif
          *argval=opts[i].val;
          return GOPT_PRESENT_WITHARG;
        }
#ifdef DEBUG
        fprintf(stderr,"...it is %s, no match.\n",opts[i].name);
#endif
        break;
      default:
#ifdef DEBUG
        fprintf(stderr,"End of options reached, returning not found.\n");
#endif
        return GOPT_NOTPRESENT;
    }
  }
}

#ifdef GOPT_FREE
/* wrapper for free() in case of several malloc implementations */
void gopt_free(void *vptr){
  free(vptr);
}
#endif
